// SPDX-FileCopyrightText: 2023 nervuri <https://nervuri.net/contact>
//
// SPDX-License-Identifier: BSD-3-Clause

// Inspired by Andrew Ayer's tlshacks:
// https://github.com/AGWA/tlshacks/blob/main/generate_extensions.go

package clienthello

import (
	_ "embed"
	"encoding/csv"
	"io"
	"log"
	"strconv"
	"strings"
)

// https://www.iana.org/assignments/tls-extensiontype-values/tls-extensiontype-values-1.csv
//
//go:embed extensions.csv
var extensionsCSV string

var ExtensionList = parseExtensionsCSV()

// === CONSTANTS ===

// TLS extension codes
const (
	extensionServerName              uint16 = 0
	extensionStatusRequest           uint16 = 5
	extensionSupportedGroups         uint16 = 10 // elliptic_curves in TLS versions prior to 1.3, see RFC 8446, Section 4.2.7
	extensionSupportedPointFormats   uint16 = 11
	extensionSignatureAlgorithms     uint16 = 13
	extensionALPN                    uint16 = 16
	extensionSCT                     uint16 = 18
	extensionPadding                 uint16 = 21
	extensionEncryptThenMac          uint16 = 22
	extensionExtendedMasterSecret    uint16 = 23
	extensionTokenBinding            uint16 = 24
	extensionCompressCertificate     uint16 = 27
	extensionRecordSizeLimit         uint16 = 28
	extensionDelegatedCredential     uint16 = 34
	extensionSessionTicket           uint16 = 35
	extensionPreSharedKey            uint16 = 41
	extensionEarlyData               uint16 = 42
	extensionSupportedVersions       uint16 = 43
	extensionCookie                  uint16 = 44
	extensionPSKModes                uint16 = 45
	extensionCertificateAuthorities  uint16 = 47
	extensionPostHandshakeAuth       uint16 = 49
	extensionSignatureAlgorithmsCert uint16 = 50
	extensionKeyShare                uint16 = 51
	extensionNextProtoNeg            uint16 = 13172 // not IANA assigned
	extensionApplicationSettings     uint16 = 17513 // not IANA assigned
	extensionChannelIDOld            uint16 = 30031 // 0x754f - not IANA assigned
	extensionChannelID               uint16 = 30032 // 0x7550 - not IANA assigned
	extensionRenegotiationInfo       uint16 = 0xff01
)

// TLS CertificateStatusType (RFC 3546)
const (
	statusTypeOCSP uint8 = 1
)

// === STRUCTS ===

// TLS 1.3 PSK Identity. Can be a Session Ticket, or a reference to a saved
// session. See RFC 8446, Section 4.2.11.
type PSKIdentity struct {
	Identity            byteSlice `json:"identity"`
	ObfuscatedTicketAge uint32    `json:"obfuscated_ticket_age"`
}

type ExtensionData struct {
	Raw                          byteSlice         `json:"raw"`
	ServerName                   string            `json:"server_name,omitempty"`
	StatusType                   uint8             `json:"status_type,omitempty"`
	SupportedGroups              []NamedGroup      `json:"supported_groups,omitempty"`
	SupportedPointFormats        []ECPointFormat   `json:"supported_point_formats,omitempty"`
	SupportedSignatureAlgorithms []SignatureScheme `json:"supported_signature_algorithms,omitempty"`
	RenegotiationInfo            []byte            `json:"renegotiation_info,omitempty"`
	ALPNProtocols                []string          `json:"alpn_protocols,omitempty"`
	SupportedVersions            []TLSVersion      `json:"supported_tls_versions,omitempty"`
	Cookie                       byteSlice         `json:"cookie,omitempty"`
	CertificateCompressionAlgos  []CertComprAlgo   `json:"compression_algorithms,omitempty"`
	KeyShares                    []KeyShare        `json:"key_shares,omitempty"`
	PSKModes                     []PSKMode         `json:"psk_modes,omitempty"`
	PSKIdentities                []PSKIdentity     `json:"psk_identities,omitempty"`
	PSKBinders                   []byteSlice       `json:"psk_binders,omitempty"`
	RecordSizeLimit              uint16            `json:"record_size_limit,omitempty"`
	Length                       uint16            `json:"length,omitempty"` // padding
}

type Extension struct {
	Code        uint16        `json:"code"`
	Name        string        `json:"name"`
	Recommended bool          `json:"-"`
	Reference   string        `json:"-"`
	Data        ExtensionData `json:"data"`
}

// === FUNCTIONS ===

func parseExtensionsCSV() map[uint16]Extension {
	extensions := map[uint16]Extension{}
	r := csv.NewReader(strings.NewReader(extensionsCSV))

	// Skip CSV header.
	_, err := r.Read()
	if err != nil {
		log.Fatal(err)
	}

	for {
		row, err := r.Read()
		if err == io.EOF {
			break
		} else if err != nil {
			log.Fatal(err)
		}
		var (
			val  = row[0]
			name = row[1]
			rec  = row[4]
			ref  = row[5]
		)
		if name == "Unassigned" {
			continue
		}
		if strings.HasPrefix(name, "Reserved") && ref != "[RFC8701]" {
			continue
		}
		code, err := strconv.Atoi(val)
		if err != nil {
			log.Println(err)
		}
		// Deal with names like "supported_groups (renamed from "elliptic_curves")"
		// by stopping at the first space.
		name = strings.SplitN(name, " ", 2)[0]
		ext := Extension{
			Code:      uint16(code),
			Name:      name,
			Reference: ref,
		}
		if rec == "Y" {
			ext.Recommended = true
		}
		if ref == "[RFC8701]" {
			ext.Name = "GREASE"
		}

		extensions[ext.Code] = ext
	}
	return extensions
}

func GetExtensionInfo(extensionCode uint16) Extension {
	info, found := ExtensionList[extensionCode]
	if !found {
		info = Extension{
			Code: extensionCode,
		}
	}
	return info
}
