// SPDX-FileCopyrightText: 2022-2023 nervuri <https://nervuri.net/contact>
//
// SPDX-License-Identifier: BSD-3-Clause

package clienthello

import "fmt"

// https://www.iana.org/assignments/tls-parameters/tls-parameters.xml#tls-parameters-9
var ECPointFormatList = map[uint8]ECPointFormatInfo{
	0: {Code: 0, Name: "uncompressed", Reference: "[RFC8422]"},
	1: {Code: 1, Name: "ansiX962_compressed_prime", Reference: "[RFC8422]"},
	2: {Code: 2, Name: "ansiX962_compressed_char2", Reference: "[RFC8422]"},
}

type ECPointFormat any // (uint8 | ECPointFormatInfo)

type ECPointFormatInfo struct {
	Code      uint8  `json:"code"`
	Name      string `json:"name"`
	Reference string `json:"-"`
}

func GetECPointFormatInfo(ecPointFormatCode uint8, mustName bool) ECPointFormatInfo {
	info, found := ECPointFormatList[ecPointFormatCode]
	if !found {
		info = ECPointFormatInfo{
			Code: ecPointFormatCode,
		}
	}
	if mustName && info.Name == "" {
		info.Name = fmt.Sprint("(", info.Code, ")")
	}
	return info
}

func (m *ClientHelloMsg) AddECPointFormatInfo() {
	for i, ext := range m.Extensions {
		if ext.Code == extensionSupportedPointFormats {
			for j, pointFormat := range ext.Data.SupportedPointFormats {
				m.Extensions[i].Data.SupportedPointFormats[j] =
					GetECPointFormatInfo(pointFormat.(uint8), false)
			}
		}
	}
}

func (m *ClientHelloMsg) GetSupportedPointFormats() []ECPointFormat {
	var supportedPointFormats []ECPointFormat
	for _, ext := range m.Extensions {
		if ext.Code == extensionSupportedPointFormats {
			supportedPointFormats = append(
				supportedPointFormats, ext.Data.SupportedPointFormats...)
		}
	}
	return supportedPointFormats
}
