// SPDX-FileCopyrightText: 2022-2023 nervuri <https://nervuri.net/contact>
//
// SPDX-License-Identifier: BSD-3-Clause

package clienthello

import (
	"fmt"
)

var TLSVersionList = map[uint16]TLSVersionInfo{
	768: {Code: 768, HexCode: "0300", Name: "SSL 3.0"},
	769: {Code: 769, HexCode: "0301", Name: "TLS 1.0"},
	770: {Code: 770, HexCode: "0302", Name: "TLS 1.1"},
	771: {Code: 771, HexCode: "0303", Name: "TLS 1.2"},
	772: {Code: 772, HexCode: "0304", Name: "TLS 1.3"},

	0x0A0A: {Code: 0x0A0A, HexCode: "0A0A", Name: "GREASE"},
	0x1A1A: {Code: 0x1A1A, HexCode: "1A1A", Name: "GREASE"},
	0x2A2A: {Code: 0x2A2A, HexCode: "2A2A", Name: "GREASE"},
	0x3A3A: {Code: 0x3A3A, HexCode: "3A3A", Name: "GREASE"},
	0x4A4A: {Code: 0x4A4A, HexCode: "4A4A", Name: "GREASE"},
	0x5A5A: {Code: 0x5A5A, HexCode: "5A5A", Name: "GREASE"},
	0x6A6A: {Code: 0x6A6A, HexCode: "6A6A", Name: "GREASE"},
	0x7A7A: {Code: 0x7A7A, HexCode: "7A7A", Name: "GREASE"},
	0x8A8A: {Code: 0x8A8A, HexCode: "8A8A", Name: "GREASE"},
	0x9A9A: {Code: 0x9A9A, HexCode: "9A9A", Name: "GREASE"},
	0xAAAA: {Code: 0xAAAA, HexCode: "AAAA", Name: "GREASE"},
	0xBABA: {Code: 0xBABA, HexCode: "BABA", Name: "GREASE"},
	0xCACA: {Code: 0xCACA, HexCode: "CACA", Name: "GREASE"},
	0xDADA: {Code: 0xDADA, HexCode: "DADA", Name: "GREASE"},
	0xEAEA: {Code: 0xEAEA, HexCode: "EAEA", Name: "GREASE"},
	0xFAFA: {Code: 0xFAFA, HexCode: "FAFA", Name: "GREASE"},
}

type TLSVersion any // (uint16 | TLSVersioninfo)

type TLSVersionInfo struct {
	Code    uint16 `json:"code"`
	HexCode string `json:"hex_code"`
	Name    string `json:"name"`
}

func GetTLSVersionInfo(tlsVersionCode uint16, mustName bool) TLSVersionInfo {
	info, found := TLSVersionList[tlsVersionCode]
	if !found {
		info = TLSVersionInfo{
			Code:    tlsVersionCode,
			HexCode: fmt.Sprintf("%04X", tlsVersionCode),
		}
	}
	if mustName && info.Name == "" {
		info.Name = "0x" + info.HexCode
	}
	return info
}

func (m *ClientHelloMsg) AddTLSVersionInfo() {
	m.RecordHeaderTLSVersion = GetTLSVersionInfo(
		m.RecordHeaderTLSVersion.(uint16), false)
	m.TLSVersion = GetTLSVersionInfo(m.TLSVersion.(uint16), false)
	for i, ext := range m.Extensions {
		if ext.Code == extensionSupportedVersions {
			for j, versionCode := range ext.Data.SupportedVersions {
				m.Extensions[i].Data.SupportedVersions[j] = GetTLSVersionInfo(
					versionCode.(uint16), false)
			}
		}
	}
}

func (m *ClientHelloMsg) GetSupportedVersions() []TLSVersion {
	var supportedVersions []TLSVersion
	for _, ext := range m.Extensions {
		if ext.Code == extensionSupportedVersions {
			supportedVersions = append(supportedVersions, ext.Data.SupportedVersions...)
		}
	}
	if len(supportedVersions) == 0 {
		// Fall back to version field.
		supportedVersions = append(supportedVersions, m.TLSVersion)
	}
	return supportedVersions
}
